import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Menu;
import java.awt.MenuBar;
import java.awt.MenuItem;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.awt.image.PixelGrabber;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;
import javax.swing.JFrame;

public class Main4 extends JFrame implements ActionListener {
	
	private static final long serialVersionUID = 1L;

	public static void main ( String[] args ) {
		Main4 main = new Main4();
		main.setVisible( true );
	}

	private DoubleBufferedPanel mainPanel;

	private File imgFile;
	private Image original;
	
	protected int RGB[][][] = new int [ 4 ][ 512 ][ 512 ];
	private double Intensities[] = new double [3];
	private double Entropies[] = new double [3];
	private int Probs[][] = new int [3] [256];
	private final double total = 512 * 512;
	private final double Dmax = 362.04;
	private final int R = 0;
	private final int G = 1;
	private final int B = 2;
	private final int DISTANCE = 3;
	private final int DevScale = 30;	// starting this small = not too much pixel randomness. change as necessary.
	private final int DistScale = 1;	// starting this big = lots of mean-scaling (not too rapid change).
										// reduce as necessary. (maybe make these cmd-line or GUI)
	
	private double threshold = 0.975;
	
	public Main4 ( ) {
		this.setSize( 550, 590 );
		initialize();
	}
	
	public Main4 ( int width, int height ) {
		this.setSize( width, height );
		initialize();
	}
	
	private void initialize(){
		this.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
		this.getContentPane().setLayout( new BorderLayout() );
		
		mainPanel = new DoubleBufferedPanel();
		mainPanel.setBackground( Color.BLACK );
		this.getContentPane().add( mainPanel, BorderLayout.CENTER );
		mainPanel.setLayout( new BorderLayout() );

		createMenus();
	}
	
	private void createMenus() {
		
		MenuBar menubar = new MenuBar();
		
		Menu file = new Menu ( "File" );
		MenuItem open = new MenuItem( "Open..." );
		open.addActionListener( this );
		file.add( open );
		
		file.addSeparator();
		
		MenuItem exit = new MenuItem ( "Exit" );
		exit.addActionListener( this );
		file.add ( exit );
		
		menubar.add( file );
		this.setMenuBar( menubar );
	}
	
	private void convertImage(){

		if ( original == null ) {
			// TODO: Notify
			return;
		}
		
		PixelGrabber grabber = new PixelGrabber( original, 0, 0, 512, 512, true );
		
		try {
			grabber.grabPixels();
		} catch (InterruptedException e) {
			// TODO: Notify
			return;
		}
		
		Object thepixels = grabber.getPixels();
		
		int[] i_buffer = null;
		
		if ( thepixels instanceof int[] )
			i_buffer = (int[]) thepixels;
		else {
			// TODO: Notify
			return;
		}
		
		int tempX, tempY;
		int r, g, b;
		
		for ( int i = 0; i < total; i++ ) {
			
			tempX = i % 512;
			tempY = i / 512;
			
			r = i_buffer[ i ] >> 16 & 0xff;
			g = i_buffer[ i ] >>  8 & 0xff;
			b = i_buffer[ i ]       & 0xff;
			
			RGB[ R ][ tempX ][ tempY ] = r;
			RGB[ G ][ tempX ][ tempY ] = g;
			RGB[ B ][ tempX ][ tempY ] = b;
			RGB[ DISTANCE ][ tempX ][ tempY ] = 
				(int)Math.sqrt( Math.pow( tempX - 256, 2 ) + Math.pow( tempY - 256, 2 ) );
			
		}
	}

	private void displayImage() {
		
		try {
			original = ImageIO.read( imgFile );
			original = original.getScaledInstance( 512, 512, 0 );
			
			mainPanel.setImage( original );
			mainPanel.paintImmediately( 0, 0, mainPanel.getWidth(), mainPanel.getHeight() );
			
			//iteration = 0;
			convertImage();
			
			while ( CalculateDifference() < threshold ) {
				/*try {
					Thread.sleep( 100 );
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}*/
				mainPanel.setImage(  TweakImage() );
				mainPanel.paintImmediately( 0, 0, mainPanel.getWidth(), mainPanel.getHeight() );
			}
			
		} catch (IOException e) {
			// TODO: Notify
			e.printStackTrace();
		}
	}
	
	public void actionPerformed(ActionEvent e) {
		
		if ( e.getActionCommand().equals( "Exit" ) ) {
			this.dispose();
		}
		else if ( e.getActionCommand().equals( "Open..." ) ) {

			JFileChooser chooser = new JFileChooser();
			int ret = chooser.showOpenDialog( this );
			
			if ( ret == JFileChooser.APPROVE_OPTION ) {
				imgFile = chooser.getSelectedFile();
				displayImage();
			}
		}

		
	}
	
	private Image TweakImage() {
		
		// color mean finds average color value as distance from center value
		double Rmean = Intensities[ R ] / total - 128;
		double Gmean = Intensities[ G ] / total - 128;
		double Bmean = Intensities[ B ] / total - 128;

		// deviation is a function of (entropy / desired entropy)^2, scaled for usability
		double Rdev = Math.pow( Entropies[ R ] / threshold, 2 ) * DevScale;
		double Gdev = Math.pow( Entropies[ G ] / threshold, 2 ) * DevScale;
		double Bdev = Math.pow( Entropies[ B ] / threshold, 2 ) * DevScale;
		
		System.out.println( "Rmean = " + Rmean + ", Gmean = " + 
				Gmean + ", Bmean = " + Bmean ); 
		System.out.println( "Rdev = " + Rdev + ", Gdev = " + 
				Gdev + ", Bdev = " + Bdev ); 
		System.out.println();
		
		double tempDist = 0.0;
	
		for ( int x = 0; x < 512; x++ )
			for ( int y = 0; y < 512; y++ ) {
				// distance from ctr, scaled for usability: farthest pixels are nearly 1/scale,
				// while closest to center are (1/Dmax)/scale
				tempDist = ( RGB[ DISTANCE ][ x ][ y ] / Dmax ) / DistScale; 
				
				// NOTE: each color is shifted toward (and maybe past) the center 
				// by (color mean +/- deviation)*distScale
				RGB[ R ][ x ][ y ] -= tempDist * (Rmean + ( Math.random() - 0.5 ) * Rdev);
				RGB[ R ][ x ][ y ] = fix ( RGB[ R ][ x ][ y ] );
				
				RGB[ G ][ x ][ y ] -= tempDist * (Gmean + ( Math.random() - 0.5 ) * Gdev);
				RGB[ G ][ x ][ y ] = fix( RGB[ G ][ x ][ y ] );
				
				RGB[ B ][ x ][ y ] -= tempDist * (Bmean + ( Math.random() - 0.5 ) * Bdev);
				RGB[ B ][ x ][ y ] = fix( RGB[ B ][ x ][ y ] );					
				
			}
		
		Image temp = UpdateImage();
		
		//iteration++;
		
		return temp;
	}

	private int fix ( int val ) {
		
		if ( val < 0 ) return Math.abs(val);
		else if ( val > 255 ) return 255 - (val % 255);
		else return val;
		
	}
	
	private BufferedImage UpdateImage() {
		BufferedImage updated = new BufferedImage ( 512, 512, BufferedImage.TYPE_INT_RGB );
		
		Graphics g = updated.getGraphics();
		
		for ( int x = 0; x < 512; x++ )
			for ( int y = 0; y < 512; y++ ) {
				g.setColor( new Color ( 
						RGB[ R ][ x ][ y ], 
						RGB[ G ][ x ][ y ], 
						RGB[ B ][ x ][ y ] ) );
				g.drawRect( x, y, 1, 1 );
			}
		
		g.dispose();
		
		/*try {
			ImageIO.write( updated, "jpeg", new File ( "temp" + iteration + ".jpeg" ) );
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}*/
		
		return updated;
	}
	
	private double CalculateEntropy() {
		
		int r, g, b;
		double colorEntropy;
		
		Intensities = new double[ 3 ];
		Probs = new int[ 3 ][ 256 ];
		Entropies = new double[ 3 ];
		
		for ( int x = 0; x < 512; x++ ) 
			for ( int y = 0; y < 512; y++ ) {
				
				r = RGB[ R ][ x ][ y ];
				g = RGB[ G ][ x ][ y ];
				b = RGB[ B ][ x ][ y ];
				
				Intensities[ R ] += r;
				Intensities[ G ] += g;
				Intensities[ B ] += b;
				
				Probs[ R ][ r ]++;
				Probs[ G ][ g ]++;
				Probs[ B ][ b ]++;
				
			}		
		
		for ( int i = 0; i < 256; i++ ) {
			
			r = Probs[ R ][ i ];
			g = Probs[ G ][ i ];
			b = Probs[ B ][ i ];
			
			if ( r != 0 )
				Entropies[ R ] += ( r / total ) *
					Math.log( r / total ) / Math.log( 2 );
			if ( g != 0 )
				Entropies[ G ] += ( g / total ) * 
					Math.log( g / total ) / Math.log( 2 );
			if ( b != 0 )
				Entropies[ B ] += ( b / total ) * 
					Math.log( b / total ) / Math.log( 2 );
		}
			
		double intensitySum = Intensities[ R ] + 
			Intensities[ G ] +
			Intensities[ B ];
		double intensityAvg = (intensitySum) / 3.0;
		
		//if ( ( Intensities[ R ] * 
		
		double rProb = Intensities[ R ] / intensitySum;
		double gProb = Intensities[ G ] / intensitySum;
		double bProb = Intensities[ B ] / intensitySum;
		
		colorEntropy = ( rProb ) * (Math.log( rProb ) / Math.log( 2 ));
		colorEntropy += ( gProb ) * (Math.log( gProb ) / Math.log( 2 ));
		colorEntropy += ( bProb ) * (Math.log( bProb ) / Math.log( 2 ));
		
		Entropies[ R ] /= -8;
		Entropies[ G ] /= -8;
		Entropies[ B ] /= -8;
		colorEntropy /= -1.6245;
		
		System.out.println( "R = " + Intensities[ R ] + ", G = " + 
				Intensities[ G ] + ", B = " + Intensities[ B ] ); 
		System.out.println( "Er = " + Entropies[R] + ", Eg = " + 
				Entropies[G] + ", Eb = " + Entropies[B] ); 
		System.out.println( "rProb = " + rProb + ", gProb = " + 
				gProb + ", bProb = " + bProb ); 
		System.out.println( "Color entropy: " + colorEntropy +  " Color Avg:" + intensityAvg);
		
		
		
		return colorEntropy;
	}
	
	private double CalculateDifference() {
		
		int r, g, b;
		double colorEntropy;
		
		Intensities = new double[ 3 ];
		Probs = new int[ 3 ][ 256 ];
		Entropies = new double[ 3 ];
		
		for ( int x = 0; x < 512; x++ ) 
			for ( int y = 0; y < 512; y++ ) {
				
				r = RGB[ R ][ x ][ y ];
				g = RGB[ G ][ x ][ y ];
				b = RGB[ B ][ x ][ y ];
				
				Intensities[ R ] += r;
				Intensities[ G ] += g;
				Intensities[ B ] += b;
				
				Probs[ R ][ r ]++;
				Probs[ G ][ g ]++;
				Probs[ B ][ b ]++;
				
			}		
		
		for ( int i = 0; i < 256; i++ ) {
			
			r = Probs[ R ][ i ];
			g = Probs[ G ][ i ];
			b = Probs[ B ][ i ];
			
			if ( r != 0 )
				Entropies[ R ] += ( r / total ) *
					Math.log( r / total ) / Math.log( 2 );
			if ( g != 0 )
				Entropies[ G ] += ( g / total ) * 
					Math.log( g / total ) / Math.log( 2 );
			if ( b != 0 )
				Entropies[ B ] += ( b / total ) * 
					Math.log( b / total ) / Math.log( 2 );
		}
			
		double intensitySum = Intensities[ R ] + 
			Intensities[ G ] +
			Intensities[ B ];
		double intensityAvg = (intensitySum) / 3.0;
		
		//if ( ( Intensities[ R ] * 
		
		double rProb = Intensities[ R ] / intensitySum;
		double gProb = Intensities[ G ] / intensitySum;
		double bProb = Intensities[ B ] / intensitySum;
		
		colorEntropy = ( rProb ) * (Math.log( rProb ) / Math.log( 2 ));
		colorEntropy += ( gProb ) * (Math.log( gProb ) / Math.log( 2 ));
		colorEntropy += ( bProb ) * (Math.log( bProb ) / Math.log( 2 ));
		
		
		
		
		Entropies[ R ] /= -8;
		Entropies[ G ] /= -8;
		Entropies[ B ] /= -8;
		colorEntropy /= -1.6245;
		
		System.out.println( "R = " + Intensities[ R ] + ", G = " + 
		Intensities[ G ] + ", B = " + Intensities[ B ] ); 
System.out.println( "Er = " + Entropies[R] + ", Eg = " + 
		Entropies[G] + ", Eb = " + Entropies[B] ); 
System.out.println( "rProb = " + rProb + ", gProb = " + 
		gProb + ", bProb = " + bProb ); 
System.out.println( "Color entropy: " + colorEntropy +  " Color Avg:" + intensityAvg);

		
		return Math.min( Intensities[ R ], Math.min( Intensities[ G ], Intensities[ B ] ) ) /
		Math.max( Intensities[ R ], Math.min( Intensities[ G ], Intensities[ B ] ) );
		
		
		
		//return colorEntropy;
	}
}
